//
//  Copyright (C) 1997,1998,1999 Forest Edge Software, All Rights Reserved
//
//  Program:    eVAX, a "Virtual VAX" for Mac OS Computers
//
//  Author:     Tom Cole
//
//  Module:     asm_pseudo.c
//
//  Purpose:    Pseudo instruction handler for the assembler.  This handles all
//              the dot-format instructions.  These instructions may store data
//              in memory, or affect the state of the virtual VAX.  For example,
//              some are used to set processor register values.
//
//  History:    07/28/97    New header format standardization
//
//              07/24/97    Converted storage pseudo ops to use store_memory() so
//                          they will be ready for virtual memory
//
//              09/15/98    Added .F_FLOAT and .D_FLOAT statements for creating
//                          constants in-memory.
//
//              12/28/98    Mark symbols created by .ENTRY with the SYM_ENTRY
//                          flag, for disassembly purposes.
//
//              01/08/99    Fix goofup with .ASCID format
//
//	         	01/11/99    Clear or fix temporary symbols when we get to
//                          a new .entry, or change the .base register, or
//                          .end an assembly.
//
//              01/18/99    When a .END explicitly references a starting
//                          address, set a flag in the assembler control
//                          block that causes the console handler to call
//                          the __ENTRY location when the assembly finishes.
//
//                          Also, mark the .ENTRY name as having to be a
//                          unique symbol definition.
//
//              01/26/99    Add .CASE support.  Also, fix .SET so that it
//                          can take an expression.
//
//              02/03/99    Add .SCB pseudo opcode, which sets SCB vector
//                          entries.  Also added .REGION operator that
//                          switches between P0 and S0 deposit locations.
//
//              02/15/99    Fixed .SCB to require aligned data, also allow
//                          stack to be specified separately.  Also added
//                          .MODE to set the mode for assembly.  Added
//                          .VECTOR as shortcut for .SCB and handler.
//
//				05/20/99	Added .MODE, .VECTOR, .CONSOLE, .INCLUDE and .IF
//
//				05/25/99	Added qualifiers to .SET for /PERM, /ENTRY, /LABEL,
//							to set symbol characteristics.  The /PERM is used
//							in ssdef.asm, for example to make status codes
//							permanently defined.
//

#include "vax.pch"
#include <ctype.h>
#include "asmproto.h"
#include "fpu.h"
#include "console_proto.h"
#include "pte.h"

//  Mapping of privileged register names

const char * pr_names[] = {
/*  0           1           2           3           4           5           6           7 */
    "KSP",      "ESP",      "SSP",      "USP",      "ISP",      "_PR05",    "_PR06",    "_PR07",
    "P0BR",     "P0LR",     "P1BR",     "P1LR",     "SBR",      "SLR",      "_PR14",    "_PR15",
    "PCBB",     "SCBB",     "IPL",      "ASTLVL",   "SIRR",     "SISR",     "_PR22",    "_PR23",
    "ICCS",     "NICR",     "ICR",      "TODR",     "_PR28",    "_PR29",    "_PR30",    "_PR31",
    "RXCS",     "RXDB",     "TXCS",     "TXDB",     "TBDR",     "_PR37",    "_PR38",    "_PR39",
    "_PR40",    "_PR41",    "_PR42",    "_PR43",    "_PR44",    "_PR45",    "_PR46",    "_PR47",
    "_PR48",    "_PR49",    "_PR50",    "_PR51",    "_PR52",    "_PR53",    "_PR54",    "_PR55",
    "MAPEN",    "TBIA",     "TBIS",     "_PR59",    "_PR60",    "PMR",      "SID",      "TBCHK",
    ""     };  /* Empty string needed by asm_expr3() */

/*  0           1           2           3           4           5           6           7 */
    

    static struct PSEUDOS {
        char * name;
        short   code;
    } pseudos[] = {
        {   "BYTE",     1   },      //  Declare a byte
        {   "DECB",     1   },
        {   "WORD",     2   },      //  Declare a word
        {   "DECW",     2   },
        {   "LONG",     3   },      //  Declare a longword
        {   "DECL",     3   },
        {   "BASE",     4   },      //  Set new base address (sets PC value)
        {   "SET",      5   },      //  Set a symbol
        {   "CLEAR",    6   },      //  Clear a symbol or all symbols
        {   "ASCII",    7   },      //  Ascii character string
        {   "ASCIZ",    8   },      //  Ascii character string followed by null byte
        {   "ASCIC",    9   },      //  Ascii character string preceded by 16-bit count
        {   "ASCID",    10  },      //  Ascii character string preceded by string descriptor
        {   "END",      11  },      //  End of input string
        {   "PSL",      12  },      //  Set PSL longword
        {   "PRINT",    13  },      //  Print a string to the console/listing
        {   "MASK"  ,   14  },      //  Create an entry point register mask
        {   "PTE",      15  },      //  Define a page table entry longword
        {   "F_FLOAT",  16  },      //  Store an F_FLOAT value
        {   "D_FLOAT",  17  },      //  Store a  D_FLOAT value
        
        //  Do not renumber the BLK* operators without modifying the code for
        //  them since they depend on the code to create a size scale value.
        
        {   "BLKB",     18  },      //  A block of bytes
        {   "BLKW",     19  },      //  A block of words
        {   "BLKL",     20  },      //  A block of longs
        {   "BLKF",     20  },      //  A block of F_FLOAT (same as longs)
        {   "BLKD",     21  },      //  A block of D_FLOATs
        
        //  Resume "free form" table assignments here.
        
        {   "ENTRY",    22  },      //  Define entry symbol AND register mask
        {   "SYM",      23  },      //  Define a system symbol

        //  Do not change the number for CASE without modifying the code
        //  below that empties the case_base field.

        {   "CASE",     24  },      //  CASEx table entry
        {   "SCB",      25  },      //  SCB vector slot
        {   "ALIGN",    26  },      //  Align deposit
        {   "REGION",   27  },      //  Choose region deposit
        {   "MODE",     28  },      //  Choose processor mode
        {   "VECTOR",   29  },      //  Declare SCB handler
        {   "CONSOLE",  30  },      //  Execute a console command
		{   "INCLUDE",  31  },      //  Include a nested assembly file
		{   "IF",       32  },      //  Conditional assembly
    
        //  End of table marker.  Must be last item.
        {   "",         -1  }};
        
    static struct RNAMES {
        char * name;
        short   number;
    } rname[] = {
        {   "R0",   0   },
        {   "R1",   1   },
        {   "R2",   2   },
        {   "R3",   3   },
        {   "R4",   4   },
        {   "R5",   5   },
        {   "R6",   6   },
        {   "R7",   7   },
        {   "R8",   8   },
        {   "R9",   9   },
        {   "R10",  10  },
        {   "R11",  11  },
        {   "R12",  12  },
        {   "R13",  13  },
        {   "R14",  14  },
        {   "R15",  15  },
        {   ".",    15  },
        {   "AP",   12  },
        {   "FP",   13  },
        {   "SP",   14  },
        {   "PC",   15  },
        {   0L, 0 }};
        

//  Assemble a pseudo operand

long asm_pseudo( struct VAX * vax, char * Buffer )
{
    char q;
    long count_pc, string_pc;
    short count;
    long oldrgn, newrgn;    
    long n, match, rc, code, perm;
    long long1, long2, scale, verb, verb2, vector, stack;
    
    union VALUE {
        unsigned char   byte[ 4 ];
        unsigned short  word[ 2 ];
        unsigned long   longword;
    } value;
    char * p = Buffer;
    char b[ 32 ], *bp, ch;

    union PTE pte;
    
    //  See if the first character is a dot... if not, then this isn't
    //  a pseudo operand and we're done.  Nonzero return means not handled.
    
    flush_blanks( &p );
    if (*(p++) != '.' )     // Skip past it while we're at it...
        return VAX_ASMNOTPSEUDO;
    
    //  Copy the opcode to a local buffer.
    
    n = 0;
    while( !isend( vax, *p ) && !is_blank( *p ) && ( *p != '/' )) {
    
        ch = *(p++);
        if( ch >= 'a' && ch <= 'z' )
            ch = ch - 32;
        b[ n++ ] = ch;
        
    }
    b[ n ] = 0;

    //  Search the list of privileged registers to see if it's a match
    
    match = 0;
    for( n = 0; n < MAXPRIVREG; n++ ) {
    
        if( strcmp( b, pr_names[ n ] ) == 0 ) {
            match = 1;
            break;
        }
    }

    //  If a match was found, it's a pseudo operand like .KSP and we set the
    //  associated privilege register now and we're done.
        
    if( match ) {

        vax-> assembler.case_base = 0L;
    
        //  Parse the actual numeric value to be placed in the register
        
        rc = asm_expr( vax, &p, ( long * ) &value.longword );
        if( rc )
            return rc;

        if( vax-> pslw.cur_mod > 0 ) {
            set_fault( vax, EXC_PRIV, 0 );
            return handle_fault( vax );
        }
                
        vax-> preg[ n ] = value.longword;
        return VAX_OK;
    }

    //  It wasn't a privileged register name, so search the list of known pseudo
    //  opcodes to find a match.
    
    match = 0;
    for( n = 0; pseudos[ n ].name[ 0 ]; n++ ) {
        if( strcmp( pseudos[ n ].name, b ) == 0 ) {
            match = 1;
            break;
        }
    }
    
    if( !match )
        return VAX_ASMNOTPSEUDO;    // Non zero code means no pseudo operand handled.
        
    //  Else handle the actual function of the routine.  If it's other
    //  than the .CASE operator, empty out the case base value so it
    //  resets the next time we do a .CASE block.
    
    code = pseudos[ n ].code;
    rc = VAX_OK;

    if( code != 24 ) /****MUST MATCH THE .CASE VALUE ****/
        vax-> assembler.case_base = 0L;
    
    switch( code ) {
    
    case 1:     //  .BYTE
        
        while( 1 ) {
        
            flush_blanks( &p );
            if( isend( vax, *p ))
                break;
            if( *p == ',' ) {
                p++;
            }
            
            rc = asm_value( vax, &p, &value.longword, K_ADDR_B );
            if( rc )
                return rc;
            
            if( value.longword < 0 || value.longword > 0x00FF )
                return VAX_ASMINVBYTE;

#if BIGENDIAN            
            rc = store_memory( vax, vax-> console.deposit, 
                          ( void * ) &value.byte[ 3 ], 1 );
#else
            rc = store_memory( vax, vax-> console.deposit, 
                          ( void * ) &value.byte[ 0 ], 1 );
#endif

            if( rc )
                return rc;

            vax-> console.deposit += 1;
        }
        
        break;
    
    case 2:     //  .WORD
        
        while( 1 ) {
        
            flush_blanks( &p );
            if( isend( vax, *p ))
                break;
            if( *p == ',' ) {
                p++;
            }
            rc = asm_value( vax, &p, &value.longword, K_ADDR_W );
            if( rc )
                return rc;
            
            if( value.longword < 0 || value.longword > 0x0000FFFFL )
                return VAX_ASMINVWORD;

#if BIGENDIAN            
            rc = store_memory( vax, vax-> console.deposit, 
                        ( void * ) &value.word[ 1 ], 2 );
#else
            rc = store_memory( vax, vax-> console.deposit, 
                        ( void * ) &value.word[ 0 ], 2 );
#endif
            if( rc )
                return rc;

            vax-> console.deposit += 2;
        }
        break;
        
    case 3:     //  .LONG

        while( 1 ) {        
            flush_blanks( &p );
            if( isend( vax, *p ))
                break;
            if( *p == ',' ) {
                p++;
            }
            rc = asm_value( vax, &p, &value.longword, K_ADDR_L );
            if( rc )
                return rc;
            
            rc = store_memory( vax, vax-> console.deposit, ( void * ) &value.longword, 4 );
            if( rc )
                return rc;

            vax-> console.deposit += 4;
            
        }
        break;
            
    case 4:     //  .BASE   value           ! Set location to store into.
                //                          ! .SET PC value     would be identical

        //  Because we're staring new, flush or fix temp symbols

        rc = scope_symbols( vax );
        if( rc )
            return rc;

        //  Get the new base value
        
        rc = asm_expr( vax, &p, ( long * ) &value.longword );
        if( rc )
            return rc;
        
        vax-> console.deposit = value.longword;
        
        break;

    case 5:     //  .SET SYM VALUE
        flush_blanks( &p );

		perm = 0;

		//  Look for qualifiers that affect the symbol.

		/*	/PERMANENT		the symbol cannot be deleted	*/

		bp = p;
		rc = read_verb( vax, &p, &verb2 );
		if( rc )
			return rc;
		if( verb2 == CHAR4( '/','P','E','R' ) ||
			verb2 == CHAR4( '/','P','R','M' )) {
			perm = perm | SYM_PERMANENT;
		}
		else {
			p = bp;
		}
		flush_blanks( &p );


		/*	/ENTRY			the symbol is an entry			*/

		bp = p;
		rc = read_verb( vax, &p, &verb2 );
		if( rc )
			return rc;
		if( verb2 == CHAR4( '/','E','N','T' )) {
			perm = perm | SYM_ENTRY;
		}
		else {
			p = bp;
		}
		flush_blanks( &p );


		/*	/LABEL			the symbol is a label			*/

		bp = p;
		rc = read_verb( vax, &p, &verb2 );
		if( rc )
			return rc;
		if( verb2 == CHAR4( '/','L','A','B' ) ||
			verb2 == CHAR4( '/','L','B','L' )) {
			perm = perm | SYM_LABEL;
		}
		else {
			p = bp;
		}
		flush_blanks( &p );


        //  Copy the name to a local buffer.
        

        n = 0;
        while( !is_blank( *p ) && !(*p == '=' ) && !isend( vax, *p )) {
        
            ch = *(p++);
            if( ch >= 'a' && ch <= 'z' )
                ch = ch - 32;
            b[ n++ ] = ch;
            
        }
        b[ n ] = 0;

        flush_blanks( &p );
        if( *p == '=' ) {
            p++;
            flush_blanks( &p );
        }
        
        rc = asm_expr( vax, &p, ( long * ) &value.longword );
        if( rc )
            return rc;
    
        //  Check for special cases of VAX processor registers.
        
        for( n = 0; rname[ n ].name; n++ ) {
            if( strcmp( b, rname[ n ].name ) == 0 ) {
                vax-> reg[ rname[ n ].number ] = value.longword;
                return VAX_OK;
            }
        }
        
        //  Otherwise, store the symbol away.  If there are extended
		//	attributes, set them as well.
        
        bp = b;
        rc = set_symbol( vax, &bp, value.longword );
		if(( rc == VAX_OK ) && perm )
			vax-> console.last_symbol-> flags |= perm;

        return rc;
    
    case 6:     //  .CLEAR SYM    or   .CLEAR ! all

        flush_blanks( &p );

        //  See if there is a name.  If not, then we clear all symbols
        
        if( isend( vax, *p ))
            return clear_all_symbols( vax );
        
        //  Copy the name to a local buffer.
        
        n = 0;
        while( !is_blank( *p ) && !isend( vax, *p )) {
        
            ch = *(p++);
            if( ch >= 'a' && ch <= 'z' )
                ch = ch - 32;
            b[ n++ ] = ch;
            
        }
        b[ n ] = 0;

                
        rc = clear_symbol( vax, b );
        return rc;
    
    case 7:     //  .ASCII  "string" [,"string"...]         ! Store characters as-is    
    case 8:     //  .ASCIZ  "string" [,"string"...]         ! Store chars followed by zero
    case 9:     //  .ASCIC  "string" [,"string"...]         ! Store counted chars
    case 10:    //  .ASCID  "string" [,"string"...]         ! Descriptor and string storage
    
    
        count = 0;
        count_pc = vax-> console.deposit;
    
        if( code == 9 ) {   /* .ascic counted string, leading word is size */
            rc = store_memory( vax, count_pc, (void * ) &count, 2 );
            if( rc )
                return rc;
            vax-> console.deposit += 2;
        }
        
        if( code == 10 ) {  /* .ascid descriptor followed by string */
        
            /* Write out length of zero.  count_pc is used later to */
            /* update the string length after parsing it.           */

            rc = store_memory( vax, count_pc, (void * ) &count, 2 );
            if( rc )
                return rc;

            /* Descriptor type, 0 = string */

            rc = store_memory( vax, count_pc + 2, (void * ) &count, 2 );
            if( rc )
                return rc;
            
            /* Advance to second longword which contains the address */
            vax-> console.deposit += 4;
            string_pc = vax-> console.deposit + 4;
            rc = store_memory( vax, vax-> console.deposit, ( void * ) &string_pc, 4 );
            if( rc )
                return rc;
            
            /* Advance past the address longword to be ready to write string */
            vax-> console.deposit += 4;
    
        }
        //  While there is more in the string...
        
        while( 1 ) {
            
            //  Find the start of the string to store in memory.  Give up if
            //  we're at the end of the string.
            
            flush_blanks( &p );
            if( isend( vax, *p ))
                break;
            
            //  See if this is a quoted string "foo", 'foo', or /foo/ and if so
            //  remember the quote and strip it.  If it isn't quoted, then just
            //  note it with a null q character.
                
            q = *p;
            if( q == '/' || q == '\'' || q == '"' )
                p++;
            else
                q = 0;
            
            //  Pick up characters from the operand buffer
            
            while( *p ) {
            
                //  If this matches the quote character we're done
                
                if( *p == q )
                    break;
                
                //  If there wasn't a quote character and this is a blank, we're done
                
                if( q == 0 && is_blank( *p ))
                    break;
                
                //  Store the character in memory and advance
                    
                rc = store_memory( vax, vax-> console.deposit++, ( unsigned char * ) p, 1 );
                if( rc )
                    return rc;
                count++;
                p++;
            }
            
            //  If the current character matches the quote (and isn't zero) then skip
            //  past the closing quote character.
            
            if (*p == q && q )
                p++;
            
            flush_blanks( &p );
            if( *p != ',' )
                break;
                
        }
        
        if( code == 8 ) {   /*  .ASCIZ */
            count = 0;
            rc = store_memory( vax, vax-> console.deposit++, ( unsigned char * ) &count, 1 );
            if( rc )
                return rc;
        }
        else
        if( code == 9 || code == 10 ) { /*  .ASCIC  or .ASCID */
            rc = store_memory( vax, count_pc, ( unsigned char * ) &count, 2 );
            if( rc )
                return rc;
        }
        
        break;  /* All done with ASCII variations... */


    case 11:    //  .END
    
        //  Fix the temp symbols, if any

        rc = scope_symbols( vax );
        if( rc )
            return rc;

        //  If there is a default entry name, set __ENTRY to that value.

        flush_blanks( &p );
        if( !isend( vax, *p )) {
        
            rc = asm_expr( vax, &p, ( long * ) &value.longword );
            if( rc )
                return rc;

            strcpy( b, "__ENTRY" );
            bp = b;
            rc = set_symbol( vax, &bp, value.longword );

            vax-> assembler.flags |= ASM_ENTRY;
        }

                
        //  Get out of assembler mode.

        vax-> console.assembler_mode = 0;
        break;

    case 12:    //  .PSL
        rc = asm_expr( vax, &p, ( long * ) &value.longword );
        if( rc )
            break;
        vax-> psl.reg = value.longword;
        
        /* Since the PSL is stored two ways, let's read the bit pattern */
        /* and update the "Wide" version too */
        
        read_psl_bits( vax );
    
        break;

    case 13:    //  .PRINT

        rc = console_print( &vax, &p );
        break;
        
    case 14:    //  .MASK
        flush_blanks( &p );
        rc = asm_mask( vax, &p, ( long * ) &value.longword );
        
        count = ( short ) value.longword;
        count_pc = vax-> console.deposit;
        
        rc = store_memory( vax, count_pc, (void * ) &count, 2 );
        if( rc )
            return rc;

        vax-> console.deposit += 2;
        break;
    
    case 15:    //  .PTE    b_valid, w_protmask, w_modifymask, w_owner, l_pfn
    
        flush_blanks( &p );
        if( *p != '<' ) {
            rc = VAX_ASMINVPTE;
            break;
        }
        pte.longword = 0L;
        
        //  Parse the valid bit value
        
        p++;
        flush_blanks( &p );
        rc = asm_expr( vax, &p, ( long * ) &value.longword );
        if( rc )
            break;
        
        pte.bit.v = value.longword;

        flush_blanks( &p );
        if( *p != ',' ) {
            rc = VAX_ASMINVPTE;
            break;
        }

        //  Parse the protection code PTE$K_xxx
            
        p++;
        flush_blanks( &p );
        rc = asm_expr( vax, &p, ( long * ) &value.longword );
        if( rc )
            break;
        
        pte.bit.prot = value.longword;

        flush_blanks( &p );
        if( *p != ',' ) {
            rc = VAX_ASMINVPTE;
            break;
        }

        //  Parse the modify bit
        
        p++;
        flush_blanks( &p );
        rc = asm_expr( vax, &p, ( long * ) &value.longword );
        if( rc )
            break;
        
        pte.bit.m = value.longword;

        flush_blanks( &p );
        if( *p != ',' ) {
            rc = VAX_ASMINVPTE;
            break;
        }

        //  Parse the PFN number.  This is a physical address, which we then
        //  shift right 9 bits to make it be a page number.
        
                
        p++;
        flush_blanks( &p );
        rc = asm_expr( vax, &p, ( long * ) &value.longword );
        if( rc )
            break;
        
        pte.bit.pfn = ( value.longword >> 9 );

        flush_blanks( &p );
        if( *p != '>' ) {
            rc = VAX_ASMINVPTE;
            break;
        }
        p++;

        count_pc = vax-> console.deposit;
        rc = store_memory( vax, count_pc, (void * ) &pte.longword, 4 );
        vax-> console.deposit += 4;
        if( rc )
            return rc;
        break;

    case 16:    // .F_FLOAT
    case 17:    // .D_FLOAT
    
        while( 1 ) {        
            flush_blanks( &p );
            if( isend( vax, *p ))
                break;
            if( *p == ',' ) {
                p++;
            }
            rc = asm_float( vax, &p );
            if( rc  )
                return rc;
            
            rc = fpu_store( vax, vax-> last_float, &long1, &long2 );
            if( rc )
                return rc;
                
            if( rc = store_memory( vax, vax-> console.deposit, ( void * ) &long1, 4 ))
                return rc;
            vax-> console.deposit += 4;

            /* If D_FLOAT then store LSB mantissa as well */
            
            if( code == 17 ) {
                if( rc = store_memory( vax, vax-> console.deposit, ( void * ) &long2, 4 ))
                    return rc;
                vax-> console.deposit += 4;
            }
            
        }
        break;

    case 18:    //  .BLKB
    case 19:    //  .BLKW
    case 20:    //  .BLKL, .BLKF
    case 21:    //  .BLKD
    
        /*  Get the count of items */
    
        flush_blanks( &p );
        rc = asm_expr( vax, &p, ( long * ) &value.longword );
        if( rc )
            break;

        scale = ( code - 18 );          //  Don't change code table without fixing this!!
        scale = 1 << scale;             //  Create a multiplier 2**scale
        
        value.longword *= scale;        //  How many bytes are we talking about?
        
        /*  Zero out the storage and advance the deposit pointer */
        
        long2 = 0;
        for( n = 0; ( unsigned long ) n < value.longword; n++ ) {
            if( rc = store_memory( vax, vax-> console.deposit, ( void * ) &long2, 1 ))
                break;
            vax-> console.deposit += 1;     //  Adjust the next deposit accordingly.
        }

        break;


    case 22:    //  .ENTRY


        //  Since we're starting a new .entry, fix the temp names
        //  of the last entry, if any.

       rc =  scope_symbols( vax );
       if( rc )
           return rc;

        //  Copy the name to a local buffer.
        
        n = 0;
        flush_blanks( &p );
        while( !is_blank( *p ) && 
               ( *p == '_' || *p == '$' || isalnum( *p ))
                 && !isend( vax, *p )) {
        
            ch = *(p++);
            if( ch >= 'a' && ch <= 'z' )
                ch = ch - 32;
            b[ n++ ] = ch;
            
        }
        b[ n ] = 0;
        
        //  Define the symbol as pointing to the current address.  We must
        //  use a pointer to the buffer since the set_symbol code wants to
        //  (re-)parse the symbol name again for us, so we give them a ptr
        //  to play with.  We know we've already advanced the p ptr ourselves.
        
        bp = b;
        
        vax-> assembler.flags |= ASM_UNIQUE;
        rc = set_symbol( vax, &bp, vax-> console.deposit );
        vax-> console.last_symbol-> flags |= SYM_ENTRY;

        if( rc )
            return rc;

        if( vax-> assembler.flags & ASM_SCOPENAMES )
	    strcpy( vax-> assembler.cur_entry, b );
        else
            vax-> assembler.cur_entry[ 0 ] = 0;

        //  Now, search for the comma character.
        
        flush_blanks( &p );
        if( *p == ',' ) {
            p++;
            flush_blanks( &p );
        }

        if( !isend( vax, *p )) {
        
            //  Parse out the mask that must follow it.
         
            rc = asm_mask( vax, &p, ( long * ) &value.longword );
            if( rc )
                return rc;

        }
        else {
            value.longword = 0;
            rc = VAX_OK;
        }

        count = ( short ) value.longword;
        count_pc = vax-> console.deposit;
        
        rc = store_memory( vax, count_pc, (void * ) &count, 2 );

        vax-> console.deposit += 2;
        break;


    case 24:  // .CASE base, item, item, item

        /* If this is the first .CASE, then establish the base */

        if( vax-> assembler.case_base == 0L )
		vax-> assembler.case_base = vax-> console.deposit;

        /* Now, scan words.  They are special CASE words */

        while( 1 ) {
        
            flush_blanks( &p );
            if( isend( vax, *p ))
                break;
            if( *p == ',' ) {
                p++;
            }
            rc = asm_value( vax, &p, &value.longword, K_CASE_W );
            if( rc )
                return rc;
            
            if( value.longword < 0 || value.longword > 0x0000FFFFL )
                return VAX_ASMINVWORD;

#if BIGENDIAN            
            rc = store_memory( vax, vax-> console.deposit, 
                        ( void * ) &value.word[ 1 ], 2 );
#else
            rc = store_memory( vax, vax-> console.deposit, 
                        ( void * ) &value.word[ 0 ], 2 );
#endif
            if( rc )
                return rc;

            vax-> console.deposit += 2;
        }
        break;

    case 25:   /* .SCB code, addr, stack */

        // Get the vector code.  Usually uses the predefined symbols
        // like ESC$ARITH are used.

        rc = asm_expr( vax, &p, ( long * ) &value.longword );
        if( rc )
            return rc;

        if( value.longword < 0 || value.longword > 0x000000FF )
            return VAX_ASMINVSCB;

        if( value.longword & 0x00000003 != 0 )
            return VAX_ASMINVSCB;

        // Get the address of the value to write.  We save the deposit addr
        // and then set it up to be the SCB address so we can make forward
        // references work.  If we get an error reading the address, we've
        // got to restore the deposit field.

        long1 = vax-> console.deposit;
        vax-> console.deposit = 0x80000000 + vax-> SCBB + value.longword;

        flush_blanks( &p );
        if( *p == ',' )
           p++;

        rc = asm_value( vax, &p, ( unsigned long * ) &value.longword, K_ADDR_L );
        if( rc ) {
            vax-> console.deposit = long1;
            return rc;
        }

        // The address MUST be aligned on a quadword boundary!

        vector = ( unsigned long ) value.longword & 0x00000003UL;

        if( vector ) {
            vax-> console.deposit = long1;
            return VAX_ASMINVSCB;
        }

        // Skip the optional comma and get the stack, if any.

        flush_blanks( &p );
        if( *p == ',' )
           p++;

        if( !isend( vax, *p )) {
           rc = read_verb( vax, &p, &verb2 );
           if( rc ) {
                vax-> console.deposit = long1;
                break;
           }
           if( verb2 == CHAR4('I','S','P',' ') )
               value.longword += 1;
           else
           if( verb2 == CHAR4('K','S','P',' ') || 
               verb2 == CHAR4('U','S','P',' ') || 
	       verb2 == CHAR4('S','S','P',' ') )
               value.longword += 0;
           else {
               vax-> console.deposit = long1;
               return VAX_ASMINVSCB;
           }
        }

        // Write the value to the memory location in the SCB, and restore
        // the deposit location to what it was.

        rc = store_memory( vax, vax-> console.deposit, ( void * ) &value.longword, 4 );
        vax-> console.deposit = long1;
        break;
        

    case 26:   /* .ALIGN size */

        //  Get the size value
        
        rc = asm_expr( vax, &p, ( long * ) &value.longword );
        if( rc )
            return rc;

        //   Divide by the alignment and multiply again.  This sets
        //   n to be the aligned value LESS than the current base.
        //   If it is less (i.e. not already aligned) then add the
        //   alignment to get aligned value GREATER than base.

        n = vax-> console.deposit / value.longword;
        n = n * value.longword;
        if( ( unsigned long ) n < vax-> console.deposit )
            vax-> console.deposit = n + value.longword;
        
        break;

    case 27:   /* .REGION S0 | SYSTEM | P0 | PROCESS */

        rc = read_verb( vax, &p, &long1 );

        switch( long1 ) {

        case CHAR4('2',' ',' ',' '):
        case CHAR4('S','0',' ',' '):
        case CHAR4('S','Y','S',' '):
        case CHAR4('S','Y','S','T'):
            newrgn = 1;
            break;

        case CHAR4('0',' ',' ',' '):
        case CHAR4('P','0',' ',' '):
        case CHAR4('P','R','O','C'):
            newrgn = 0;
            break;

        default:
            return VAX_ASMINVRGN;
        }

        //  See where we are currently depositing.  If it's the same
        //  as the region we're switching to we have no work to do.

        oldrgn = ( long ) vax-> console.deposit < 0;
        if( oldrgn == newrgn )
            return VAX_OK;

        //  If we were depositing to P0 then save the current deposit
        //  location there, and swith to the saved S0 location.

        if( oldrgn == 0 ) {
            vax-> console.p0_deposit = vax-> console.deposit;
            vax-> console.deposit = vax-> console.s0_deposit;
        }

        //  Otherwise we were depositing to S0 so save the current one
        //  there and switch to the saved P0 location.

        else {
            vax-> console.s0_deposit = vax-> console.deposit;
            vax-> console.deposit = vax-> console.p0_deposit;
        }

        rc = VAX_OK;        

        break;

    case 28:   /* .MODE USER | KERNEL | SUPERVISOR | EXEC */

        rc = read_verb( vax, &p, &long1 );
        if( rc )
            break;

        switch( long1 ) {

        case CHAR4('U',' ',' ',' '):
        case CHAR4('U','S','E','R'):
            long1 = AM_USER;
            break;

        case CHAR4('S',' ',' ',' '):
        case CHAR4('S','U','P','E'):
            long1 = AM_SUPERVISOR;
            break;

        case CHAR4('E',' ',' ',' '):
        case CHAR4('E','X','E','C'):
            long1 = AM_EXECUTIVE;
            break;

        case CHAR4('K',' ',' ',' '):
        case CHAR4('K','E','R','N'):
            long1 = AM_KERNEL;
            break;

        }
        set_mode_stack( vax, long1 );
        invalidate_tb_prot();
        break;

    case 29:   /* .VECTOR code, stack */

        // Get the vector code.  Usually uses the predefined symbols
        // like ESC$ARITH are used.

        rc = asm_expr( vax, &p, ( long * ) &value.longword );
        if( rc )
            return rc;

        if( value.longword < 0 || value.longword > 0x000000FF )
            return VAX_ASMINVSCB;

        if( value.longword & 0x00000003 != 0 )
            return VAX_ASMINVSCB;

        vector = value.longword;
        long1 = vax-> console.deposit;
        vax-> console.deposit = 0x80000000 + vax-> SCBB + vector;

        flush_blanks( &p );
        if( *p == ',' )
           p++;

        // Skip the optional comma and get the stack, if any.

        flush_blanks( &p );
        if( *p == ',' )
           p++;
        stack = 0;
        if( !isend( vax, *p )) {
           rc = read_verb( vax, &p, &verb2 );
           if( rc ) {
                vax-> console.deposit = long1;
                break;
           }
           if( verb2 == CHAR4('I','S','P',' ') )
               stack = 1;
           else
           if( verb2 == CHAR4('E','S','P',' ') || verb2 == CHAR4('K','S','P',' ') || 
               verb2 == CHAR4('U','S','P',' ') || verb2 == CHAR4('S','S','P',' ') )
                   stack = 0;
           else {
               vax-> console.deposit = long1;
               return VAX_ASMINVSCB;
           }
        }

        // Get the address of the value to write.  This is the current
        // deposit location aligned on an 8 byte boundary.

        n = long1 / 8;
        n = n * 8;
        if( n < long1 )
            long1 = n + 8;

        // Write the value to the memory location in the SCB, and restore
        // the deposit location to where we start writing the handler.

        value.longword = long1 + stack;
        
        long2 = vax-> pslw.cur_mod;
        vax-> pslw.cur_mod = 0;

        rc = store_memory( vax, vax-> console.deposit, 
               ( void * ) &value.longword, 4 );

        vax-> pslw.cur_mod = long2;
        vax-> console.deposit = long1;
        break;
        
    case 30:   // .CONSOLE <cmd>

        flush_blanks( &p );
        vax-> console.assembler_mode = 0;
        rc = console_dispatch( &vax, p );
        vax-> console.assembler_mode = 1;
        break;

    case 31:   //  .INCLUDE "file.asm"

		flush_blanks( &p );
		rc = console_asm( &vax, &p );
		break;

    case 32:   //   .IF  expression <stmt>

		rc = asm_expr( vax, &p, &long1 );
        if( rc )
			return rc;

		flush_blanks( &p );
		bp = p;

		read_verb( vax, &p, &verb );
		if( verb != CHAR4('T','H','E','N') )
			p = bp;

		if( long1 == 0L ) {

			while( !isend( vax, *p ))
				p++;
		}

		return assemble( &vax, &p );

    }

    
    return rc;
    
}
