//
//  Copyright (C) 1997,1998,1999 Forest Edge Software, All Rights Reserved
//
//  Program:    eVAX, a "Virtual VAX" for Mac OS Computers
//
//  Author:     Tom Cole
//
//  Module:     decode_opcode.c
//
//  Purpose:    This module handles decoding the instruction.  This involves validating
//              and identifying the opcode, and then arranging (via decode_operand.c) for
//              the decoding of the instruction operands.  
//
//              This is used both as a precursor to instruction execution, and as the
//              mechanism for disassembling an instruction (disassembly is controlled
//              by the "disasm" flag).
//
//  History:    07/28/97    New header format standardization
//
//              12/28/98    Add support for disassembling an entry point
//                          mask word.  We determine these from looking to see
//                          if the location we are decoding matches a SYM_ENTRY
//                          symbol entry.
//
//              05/10/99    Add support for system symbol table entries in
//                          disassembly.
//

#include "vax.pch"
#include "vaxinstr.h"

#include "asmproto.h"       // for get_symbol()


//
//  Decode an instruction by addressing modes
//

long decode_instruction( struct VAX * vax, struct OPCODE * opcode, short disasm )
{
    struct SYMBOL * sp;
    short n, j, treg, opcount, found;
    unsigned char * addr;
    unsigned short mask;
    unsigned char f;
    long pc, rc;
    char opname[ 10 ];
    char * scale;
    char label[ 20 ], label2[ 20 ];
    char * disasm_buff;    
    
    vax-> instruction_PC = pc = vax-> PC;
    


    //  Set up the disassembly if we're doing that.  Check for the 
    //  possibility that this is an .ENTRY mask as a special case.

    if( disasm ) {

        addr = ( unsigned char * ) pc;
        sprintf( opcode-> name, "%08lX: ", (long) addr );

        vax-> assembler.comment[ 0 ] = 0;

        for( sp = vax-> console.symbols; sp; sp = sp-> next )
            if(( sp-> flags & SYM_ENTRY ) && sp-> value == pc )
                break;

        if( !sp )  /* Not in user table, try system */
        for( sp = vax-> console.system_symbols; sp; sp = sp-> next )
            if(( sp-> flags & SYM_ENTRY ) && sp-> value == pc )
                break;

        if( sp ) {
            rc = load_memory( vax, pc, ( void * ) &mask, 2 );
            if( rc )
                return rc;

            strcat( opcode-> name, "              .ENTRY " );
            strcat( opcode-> name, sp-> name );
            strcat( opcode-> name, ", " );
            strcat( opcode-> name, format_mask( mask ));
  
            vax-> PC = pc + 2;
            return VAX_OK;
        }

        /*  Not a .ENTRY but might be a label... */

        for( sp = vax-> console.symbols; sp; sp = sp-> next ) 
            if(( sp-> flags & SYM_LABEL ) && sp-> value == pc )
                break;

        if( !sp )  /* Not in user table, try system table */
        for( sp = vax-> console.system_symbols; sp; sp = sp-> next ) 
            if(( sp-> flags & SYM_LABEL ) && sp-> value == pc )
                break;

        if( sp ) {
            sprintf( label, "%s:", sp-> name );
            sprintf( label2, "%-14s", label );
            strcat( opcode-> name, label2 );
        }
        else
            strcat( opcode-> name, "              " );
    }
    else
        opcode-> name[ 0 ] = 0;


    
    //  Get the function opcode.  If it's a two-byte function code, then
    //  save the extended function code bits and fetch the lsb of the opcode.

    rc = load_byte( vax, pc, ( unsigned char * ) &f );
    if( rc )
        return rc;
        
    pc++;
    
    if( f > 0x00FC ) {
        opcode-> extended = f;
        rc = load_byte( vax, pc, ( unsigned char * ) &f );
        if( rc )
            return rc;
		pc++;
    }
    else {
        opcode-> extended = 0;
    }
    
    opcode-> function = f;
    
    //  If it's not a two-byte operand, just look up the count in the table
    //  of opcodes.  This works for most of the instructions (and all the
    //  common ones.
    
    if( opcode-> extended == 0 ) {
        opcount = instruction[ f ].operand_count;
        opcode-> index = f;
    }

    //  Nope, It's a two-byte one, so we must search the table starting with
    //  the extended instructions looking for a match.  The table is ended with
	//	a operand count of 0xFF.

    else {
        for( found = 0, n = 256; instruction[ n ].operand_count <= 6; n++ ) {
            if( opcode-> extended == instruction[ n ].extended && 
                f == instruction[ n ].opcode ) {
                    opcount = instruction[ n ].operand_count;
                    opcode-> index = n;
                    found = 1;
                    break;
            }
        }

        //  If we never found the instruction, then it's a privileged or
        //  reserved-to-digital fault.  It takes a count of zero arguments
        //  This call will redirect the exception to the fault handler, or
        //  will stop the processor if the exception vector is empty.

        if( !found )
            return set_fault( vax, EXC_PRIV, 0 );
    }

    
    if( disasm ) {
        sprintf( opname, "%-8s ", instruction[ opcode-> index ].name );
        strcat( opcode-> name, opname );
    }
        
    opcode-> count = ( unsigned char ) opcount;
    vax-> PC = pc;
        
    //  Record the profiling info.  Currently this just counts
    //  the number of times each instruction is used.  Note that
    //  since this code is used for disassembly as well as executing
    //  we only count if we are not halted.

    if( !vax->halted )
        instruction[ opcode-> index ].use_count++;

    //  If there are no operands, we're done!
       
    if( opcount == 0 )
        return VAX_OK;
        
    //  Nope, work to be done.  Set up the first available temp register
    //  which is used for literal handling, etc.
        
    treg = 16; 
    opcode-> access = instruction[ opcode-> index ].access;
    scale = instruction[ opcode-> index].scale;
    disasm_buff = disasm ? opcode-> name : 0L;
        
    //  Let's scan over the operands and scoop them up.


    for( j = 0; j < opcount; j++ ) {
        
        rc = decode_operand( vax, opcode, j, &pc, &treg,
                  scale[ j ], 0, disasm_buff );
            
        if( rc ) {
            vax-> PC = pc;
            return rc;
        }
            

    }   // End of operand loop
    
    vax-> PC = pc;

    //  If we're disassembling, add the comment to the output.

    if( disasm && vax-> assembler.comment[ 0 ] != 0 ) {

        pad( opcode-> name, 64 );
        strcat( opcode-> name, vax-> assembler.comment );
        vax-> assembler.comment[ 0 ] = 0;

    }
    
    return VAX_OK;
}

/*
 *  Utility routine to pad a buffer with blanks to a specific length.
 */

void pad( char * buff, int len )
{

    int size, n;

    size = strlen( buff );
    if( size > len ) {
       strcat( buff, " " );
       return;
    }

    for( n = size; n < len ;n++ ) 
        buff[ n ] = ' ';
    buff[ len ] = 0;
    return;
}


