//
//  Copyright (C) 1997,1998,1999 Forest Edge Software, All Rights Reserved
//
//  Program:    eVAX, a "Virtual VAX" for Mac OS Computers
//
//  Author:     Tom Cole
//
//  Module:     emul_misc.c
//
//  Purpose:    Emulator handlers for miscellaneous instructions
//
//  History:    07/28/97    New header format standardization
//
//				05/29/31	Added INSQHI, INSQTI instructions


#include "vax.pch"
#include "fpu.h"

//  Misc utility macro to get operands.

#define GET_OP( n, v ) \
    p32 = ( long * ) get_operand( vax, opcode, n, OP_RD );   \
    if( !p32 )                                               \
        return VAX_FAULT;                                    \
    v = *p32



//	BUGL, BUGW - bugcheck

EMULATOR_ENTRY( emul_bug )
{

	long code;

//  The BUGL/BUGW instructions takes a single implicit immediate instruction argument.

    code = opcode-> VAXaddr[ 0 ];

//	Use that to define the fault code, and we're done.

	set_fault( vax, EXC_PRIV, 1, code );

	return VAX_FAULT;
}




//  INDEX - compute an index for array references in a HLL.

EMULATOR_ENTRY( emul_index )
{

    long subscript, low, high, size, in, out;

    long *p32;

    //  Get all the various input parameters

    GET_OP( 0, subscript );
    GET_OP( 1, low );
    GET_OP( 2, high );
    GET_OP( 3, size );
    GET_OP( 4, in );

    //  Make sure the output parameter is addressable

    p32 = ( long * ) get_operand( vax, opcode, 5, OP_WR );
    if( !p32 )
        return VAX_FAULT;

    //  Calculate the index result value

    out = ( in + subscript ) * size;

    //  See if the value blows the limits of the range.  IF so,
    //  we signal a fault.

    if(( subscript < low ) || ( subscript > high )) {
        set_fault( vax, EXC_ARITH, 1, TRAP_SUB_RNG );
        return VAX_FAULT;
    }

    vax-> pslw.n = ( out <  0L );
    vax-> pslw.z = ( out == 0L );
    vax-> pslw.v = 0;
    vax-> pslw.c = 0;

    *p32 = out;
    return VAX_OK;
}


//  BISPSW, BICPSW - set or clear bits in the PSW (low-order word of PSL) under
//                   user control.

EMULATOR_ENTRY( emul_bitpsw )
{

    short mask;
    long longmask;

//  Get the mask value.

	GET_OPERAND( mask, short, 0, OP_RD );

    
//  First, if bits 8:15 are non-zero then it's a fault.

    if( mask & 0xFF00 )
        return set_fault( vax, EXC_RESOP, 0 );

//  Mask the bits to the PSL that we care about.  Depending on
//  the opcode, the mask is used to set or clear the bits...

    longmask = ( long ) mask && 0x000000FF;
    
    write_psl_bits( vax );  /* Make sure bits match cached data */
    
    if( opcode-> function == 0xB8)
        vax-> psl.reg |= longmask;      /* BISPSW */
    else
        vax-> psl.reg &= ~longmask;     /* BICPSW */
    
    return VAX_OK;
}



//  PUSHR   Use a mask word to define registers to be pushed onto the stack.  You cannot
//          specify the PC with this instruction.

EMULATOR_ENTRY( emul_pushr )
{

    short mask, *mask_a;
    long n;
    
//  Get the mask value that defines which registers to process.

    mask_a = ( short * ) get_operand( vax, opcode, 0, OP_RD );
    if( mask_a == 0L )
        return VAX_FAULT;
    mask = *mask_a;

//  Loop over the list, pushing the data

    for( n = 14; n >= 0; n-- ) {
    
        if(( mask >> n ) & 0x0001 ) {
        
            vax-> SP -= 4;
            store_memory( vax, vax-> SP, ( void * ) &( vax-> reg[ n ]), 4 );
        }
    }
    
    return VAX_OK;
}




//  POPR    Use a mask word to define registers to be removed from the stack.

EMULATOR_ENTRY( emul_popr )
{

    short mask, *mask_a;
    long n;
    
//  Get the mask value that defines which registers to process.

    mask_a = ( short * ) get_operand( vax, opcode, 0, OP_RD );
    if( mask_a == 0L )
        return VAX_FAULT;
    mask = *mask_a;

//  Loop over the list, popping the register values

    for( n = 0; n <= 14; n++ ) {
    
        if( mask & ( 1 << n )) {
            load_memory( vax, vax-> SP, ( unsigned char * ) &( vax-> reg[ n ]), 4 );
            vax-> SP += 4;
        }
    }
    
    return VAX_OK;
}








//  HALT    Halts the processor on the next instruction boundary and returns control
//          to the console.

EMULATOR_ENTRY( emul_halt )
{

    vax-> halted = VAX_HALT;
    return VAX_OK;

}




//  NOP     Does no work, as cheaply as possible.  Later, we'll short-circuit the
//          decoder to recognize NOP as a special case that should not even be dispatched.

EMULATOR_ENTRY( emul_nop )
{
    return VAX_OK;
}





//  BPT     Signal a breakpoint exception

EMULATOR_ENTRY( emul_bpt )
{
    return set_fault( vax, EXC_BPT, 0L );
}




//	MOVPSL	Write the whole PSL somewhere.

EMULATOR_ENTRY( emul_movpsl )
{

	write_psl_bits( vax );
	return put_operand( vax, opcode, 0, OP_WR, ( char * ) &vax-> psl.reg );
}




//	INSQHI

EMULATOR_ENTRY( emul_insqhi )
{


	unsigned long	entry, header, offset, hf, hb;
	long	rc;

	/*	Get the parameters */

	entry = opcode-> VAXaddr[ 0 ];
	header = opcode-> VAXaddr[ 1 ];

	/*	Get the header forward link */

	rc = load_memory( vax, header, ( void * ) &hf, 4 );
	if( rc )
		return rc;
	hf = header + hf;

	/*	Get the header backwards link */
	rc = load_memory( vax, header+4, ( void * ) &hb, 4 );
	if( rc )
		return rc;
	hb = header+ hb;

	vax-> pslw.n = 0;
	vax-> pslw.v = 0;
	vax-> pslw.c = 0;

	/*	Simplest case is that this is the first added to the queue */

	if( hf == 0 && hb == 0 ) {

		offset = entry - header;
		rc = store_memory( vax, header, ( void * ) &offset, 4 );
		if( rc )
			return rc;

		rc = store_memory( vax, header+4, ( void * ) &offset, 4 );
		if( rc )
			return rc;

		offset = header - entry;
		rc = store_memory( vax, entry, ( void * ) &offset, 4 );
		if( rc )
			return rc;

		rc = store_memory( vax, entry+4, ( void * ) &offset, 4 );
		if( rc )
			return rc;

		vax-> pslw.z = 1;
		return VAX_OK;
	}


	/* Harder case is actually adding it in to the queue */

	/* Fix current first entry's back link to be new entry */
	offset = entry - hf;
	rc = store_memory( vax, hf+4, ( void * ) &offset, 4 );
	if( rc )
		return rc;

	/* Fix new entry's back link to be the header */
	offset = header - entry;
	rc = store_memory( vax, entry+4, ( void * ) &offset, 4 );
	if( rc )
		return rc;

	/* Fix header's forward link to be the new entry. */
	offset = entry - header;
	rc = store_memory( vax, header, ( void * ) &offset, 4 );
	if( rc )
		return rc;

	/* Fix new entry's forward link to be old first entry */
	offset = hf - entry;
	rc = store_memory( vax, entry, ( void * ) &offset, 4 );
	if( rc )
		return rc;

	return VAX_OK;

}




//	INSQTI

EMULATOR_ENTRY( emul_insqti )
{


	unsigned long	entry, header, offset, hf, hb;
	long	rc;

	/*	Get the parameters */

	entry = opcode-> VAXaddr[ 0 ];
	header = opcode-> VAXaddr[ 1 ];

	/*	Get the header forward link */

	rc = load_memory( vax, header, ( void * ) &hf, 4 );
	if( rc )
		return rc;
	hf = header + hf;

	/*	Get the header backwards link */
	rc = load_memory( vax, header+4, ( void * ) &hb, 4 );
	if( rc )
		return rc;
	hb = header+ hb;

	vax-> pslw.n = 0;
	vax-> pslw.v = 0;
	vax-> pslw.c = 0;

	/*	Simplest case is that this is the first added to the queue */

	if( hf == 0 && hb == 0 ) {

		offset = entry - header;
		rc = store_memory( vax, header, ( void * ) &offset, 4 );
		if( rc )
			return rc;

		rc = store_memory( vax, header+4, ( void * ) &offset, 4 );
		if( rc )
			return rc;

		offset = header - entry;
		rc = store_memory( vax, entry, ( void * ) &offset, 4 );
		if( rc )
			return rc;

		rc = store_memory( vax, entry+4, ( void * ) &offset, 4 );
		if( rc )
			return rc;

		vax-> pslw.z = 1;
		return VAX_OK;
	}


	/* Harder case is actually adding it in to the queue */

	/* Fix current last entry's forward link to be new entry */
	offset = entry - hb;
	rc = store_memory( vax, hb+4, ( void * ) &offset, 4 );
	if( rc )
		return rc;

	/* Fix new entry's forward link to be the header */
	offset = header - entry;
	rc = store_memory( vax, entry, ( void * ) &offset, 4 );
	if( rc )
		return rc;

	/* Fix header's backward link to be the new entry. */
	offset = entry - header;
	rc = store_memory( vax, header+4, ( void * ) &offset, 4 );
	if( rc )
		return rc;

	/* Fix new entry's backward link to be old last entry */
	offset = hb - entry;
	rc = store_memory( vax, entry+4, ( void * ) &offset, 4 );
	if( rc )
		return rc;

	return VAX_OK;

}




//	INSQUE

EMULATOR_ENTRY( emul_insque )
{


	long	entry;
	long	pred;
	long	pred_forw;
	long	rc;

	/*	Get the parameters */

	entry = opcode-> VAXaddr[ 0 ];
	pred = opcode-> VAXaddr[ 1 ];

	/* Need to validate that all the memory is accessible! */
	/* TBD */


	/* Step 1.  The entry's forward link is set to pred's forward link */

	/* Get pred's forward link (pred) */
	rc = load_memory( vax, pred, ( void * ) &pred_forw, 4 );
	if( rc )
		return rc;

	/* Store in new entries forward link (entry) */
	rc = store_memory( vax, entry + 0, ( void * ) &pred_forw, 4 );
	if( rc )
		return rc;

	vax-> pslw.n = ( pred_forw < pred );
	vax-> pslw.z = ( pred_forw == pred );
	vax-> pslw.v = 0;
	vax-> pslw.c = ( (unsigned long) pred_forw < ( unsigned long ) pred );

	/* Step 2.	The entry's backward link is set to point to pred */

	/* Store pred as backward link of new entry */

	rc = store_memory( vax, entry + 4, ( void * ) &pred, 4 );


	/* Step 3.  The backward link of the successor is set to the entry */

	/* Store in back link */
	rc = store_memory( vax, pred_forw+4, ( void * ) &entry, 4 );
	if( rc )
		return rc;

	
	/* Step 4.  The forward link of the predecessor is set to the entry */

	rc = store_memory( vax, pred, ( void * ) &entry, 4 );
	if( rc )
		return rc;


	return VAX_OK;

}




//	REMQHI

EMULATOR_ENTRY( emul_remqhi )
{


	long	entry;
	long	header;

	/*	Get the parameters */

	entry = opcode-> VAXaddr[ 0 ];
	header = opcode-> VAXaddr[ 1 ];


	return VAX_OK;

}




//	REMQTI

EMULATOR_ENTRY( emul_remqti )
{


	long	entry;
	long	header;

	/*	Get the parameters */

	entry = opcode-> VAXaddr[ 0 ];
	header = opcode-> VAXaddr[ 1 ];

	return VAX_OK;

}




//	REMQUE

EMULATOR_ENTRY( emul_remque )
{


	long	entry, addr_succ, addr_pred;
	long	rc;

	/*	Get the parameters */

	entry = opcode-> VAXaddr[ 0 ];


	/* Validate that all the memory references will work! */
	/* TBD */


	/*	Step 1.	Fix forward link of our predecessor to point */
	/*			to our successor.							 */

	/*	Get address of our predecessor */

	rc = load_memory( vax, entry+4, ( void * ) &addr_pred, 4 );
	if( rc )
		return rc;

	/*	Get address of our successor */

	rc = load_memory( vax, entry,   ( void * ) &addr_succ, 4 );
	if( rc )
		return rc;

	vax-> pslw.n = ( addr_succ < addr_pred );
	vax-> pslw.z = ( addr_succ == addr_pred );
	vax-> pslw.v = ( entry == addr_succ );
	vax-> pslw.c = ( ( unsigned long ) addr_succ < ( unsigned long ) addr_pred );

	/*	Store our successor in our predecessor's forward link */

	rc = store_memory( vax, addr_pred, ( void * ) &addr_succ, 4 );
	if( rc )
		return rc;


	/*	Step 2.	Fix backward link of our successor to be	*/
	/*			our predecessor.							*/


	rc = store_memory( vax, addr_succ + 4, ( void * ) &addr_pred, 4 );
	if( rc )
		return rc;


	rc = put_operand( vax, opcode, 1, OP_WR, ( void * ) &entry );

	return VAX_OK;

}


