//
//  Module:     shim.c
//
//  Purpose:    This module contains the code that manages runtime library shims.
//				When a RUN command is used to invoke a VMS image file, it wil usually
//				contain references to runtime libraries.  These can be resolved by
//				"software shims" which represent the real runtime routines, but
//				transfer control here instead.
//
//				To create a shim, define the entry point for the routine as a standard
//				VAX routine.  Many of these are defined in the kernel.asm file, for
//				example.
//
//				Then, create a shim label, of the form SHIM$<library>_<vector> where the
//				name is a string with the library name (LIBRTL, DECC_SHR, etc.) and
//				the vector entry expressed as a hexadecimal number.  For example,
//				the LIB$PUT_OUTPUT routine is at vector offset 00000478 in the LIBRTL
//				sharable image.  So the shim would define SHIM$LIBRTL_00000478 as the
//				symbolic name of the shim, and equate it to the LIB$PUT_OUTPUT routine
//				defined in the microkernel.
//
//				Note that to ensure that SHIM names are created as system symbols, they
//				have the SHIM$ prefix added.
//
//				When a shim runs, it may want to implement the code in native VAX, as
//				in the LIB$PUT_OUTPUT entry, or may want to vector it to a native
//				runtime in the emulator.  For this purpose, a special XFC entry is
//				used, XFC XFC$SHIM
//
//				When the XFC handler sees this, it transfers control to this dispatch
//				routine.  The dispatcher uses the contents of R0 to define which run
//				time entry is to be invoked.  This is called, and can use any of the
//				information in the virtual machine state to do it's work.
//
//				When the shim handler returns, control resumes on the instruction that
//				follows the XFC call, assuming that the runtime has set any required
//				resulting state information in the machine (return codes, etc.).
//				
//
//  History:    07/14/99    Initially created.
//


#include "vax.pch"
#include "vaxinstr.h"
#include "asmproto.h"       // for get_symbol()
#include "memmap.h"  		// to instantiate memory map list headers (uses GLOBAL)
#include "shim.h"

struct RTL_ENTRY rtl_entry_list[] = {
{	0L,		0,	"<null>"		},
{	0L,		1, 	"LIB$ADAWI"		},		/*		1 */
{	0L,		1,	"STR$UPCASE"	},
{	0L,		0,	"<end-of-list>"	}};




long shim_declare(  long  code, void * entry )
{

	rtl_entry_list[ code ].addr = entry;
	
	return VAX_OK;
}

long shim_init( void )
{

	shim_declare( 1, lib_adawi );
    shim_declare( 2, str_upcase );
    
	return VAX_OK;
}





long shim( struct VAX * vax )
{
	long	code;
	long	argc;
	long	*argv;
	long	size;
	long	rc, n;
	
	long	(*callv)( struct VAX *, long, long * );
	long	(*call0)( struct VAX * );
	long	(*call1)( struct VAX *, long );
	long	(*call2)( struct VAX *, long, long );
	long	(*call3)( struct VAX *, long, long, long );
	 
	code = vax-> R0;
	
/*
 *	Find out how many arguments there are
 */

	rc = load_memory( vax, vax-> AP, ( void * ) &argc, 4 );
	if( rc )
		return rc;

/*
 *	Allocate an argument array natively for them.
 */

	size = sizeof( long ) * argc;
	argv = ( long * ) getmem( size );
	if( argv == 0L )
		return VAX_MEM;

/*
 *	Get the arguments locally in native format
 */

	for( n = 0; n < argc; n++ ) {
	
		rc = load_memory( vax, ( vax-> AP ) + (( n + 1 ) * 4 ), ( void * ) &( argv[ n ]), 4 );
		if( rc )
			return rc;
		
	}

	if( rtl_entry_list[ code ].calltype == 0 ) {
		callv = ( void * ) rtl_entry_list[ code ].addr;
		rc = (*callv)( vax, argc, argv );
	}
	else
	switch( argc ) {
	
	case 0:		call0 = ( void * ) rtl_entry_list[ code ].addr ; rc = (*call0)( vax );
				break;
	
	case 1:		call1 = ( void * ) rtl_entry_list[ code ].addr; rc = (*call1)( vax,
																			   argv[ 0 ]);
				break;

	case 2:		call2 = ( void * ) rtl_entry_list[ code ].addr; rc = (*call2)( vax,
																			   argv[ 0 ],
																		  	   argv[ 1 ]);
				break;

	case 3:		call3 = ( void * ) rtl_entry_list[ code ].addr; rc = (*call3)( vax,
																			   argv[ 0 ],
																		  	   argv[ 1 ],
																		  	   argv[ 2 ]);
				break;

	}
	
	vax-> R0 = rc;
	freemem( argv );


	return VAX_OK;
}
