//
//  Copyright (C) 1997,1998,1999 Forest Edge Software, All Rights Reserved
//
//  Program:    eVAX, a "Virtual VAX"
//
//  Author:     Tom Cole
//
//  Module:     vax.c
//
//  Purpose:    Main VAX execution controller module.  This routine handles the
//              execution of sequential VAX instructions starting at either a
//              numeric address or at a symbol.  This routine also handles support
//              for the t-bit to identify when a trace fault should occur.
//
//  History:    07/28/97    New header format standardization
//
//              05/07/99    Add the current SP (and implicitly mode) info
//                          to the trace output.
//
//				06/14/99	Added ability to set breakpoints on FAULTS as well
//							as instruction addresses.

#if macintosh
#include <SIOUX.h>
#endif

#define GLOBAL

#include "vax.pch"
#include "vaxinstr.h"
#include "asmproto.h"       // for get_symbol()
#include "memmap.h"  		// to instantiate memory map list headers (uses GLOBAL)


//  Begin execution of instructions at a particular address denoted by
//  a symbolic value.

long execute_symbol( struct VAX * vax, char * symbol, short disasm )
{

    long    pc, rc;
    char * cp;
    
    cp = symbol;
    rc = get_symbol( vax, &cp, &pc );
    if( rc )
        return rc;
    
    vax-> PC = pc;
    return execute_vax( vax, disasm );
}



//
//  Begin execution of instructions

long execute_vax( struct VAX * vax, short disasm )
{

    struct OPCODE opcode;
    long rc, saved_pc;
    long (*routine)(struct VAX * vax, struct OPCODE * opcode ); 
    struct BREAKSTR * brkp;
    unsigned long initial_PC;
    static char * mode_name[] = { "KSP", "ESP", "SSP", "USP" };
    char * modep;
	struct INTERRUPT * ip, *ipnext, *iplast;
	long iquantum, minquantum;
	
    vax-> halted = 0;
    initial_PC = vax-> PC;
    
    while( !vax-> halted ) {

		//	If we have a pending fault from a FAULT BREAKPOINT, trigger it now.
		
		if( vax-> fault_pending ) {
			rc = handle_fault( vax );
			if( rc )
				return rc;
		}
		
		//	Update the clock if it's enabled.

		if( vax-> clock_running ) {

			if( vax-> clock == 0xFFFFFFFF) {

				// If an interrupt already happened, set ICCS<ERR> bit.
				if( vax-> ICCS & ( 1<< 7)) 
					vax-> ICCS |= 0x80000000;

				// Set the interrupt bit.

				vax-> ICCS |= ( 1 << 7 );

				// If ICCS<IE> then also declare an interrupt at IPL 22

				if( vax-> ICCS & ( 1 << 6 ))
					interrupt( vax, 0xC0, 22, 0 );

				// Load up with the next interval timer value.

				vax-> clock = vax-> NICR;
			}
			else
				vax-> clock++;

		}

		//	Handle GUI issues.  On cooperative task systems like the Mac, you have to
		//	yield once in a while to let other stuff run, including handling GUI events.
		
		if( vax-> uiquantum.initial ) {
		
			vax-> uiquantum.current++;
			if( vax-> uiquantum.current > vax-> uiquantum.initial ) {
			
				vax-> uiquantum.current = 0;
				
				/* Do the actual cooperative yield */

		
#if macintosh
        //  If we are allowing the emulator to time slice, check for it now.  On
        //  "quantum" boundaries of instruction executions, we allow the native
        //  console handler (SIOUX, part of CodeWarrior) to handle an event that
        //  it grabs from the event queue.  We call it until it's got all the
        //  events under control again.
        
                while( SIOUXHandleOneEvent(0L));
#endif

			}
		}
		
		//	Handle the interrupts.  We don't do this if quantum is turned off.  Don't
		//	turn this off unless you really want to prevent interrupt handling.  You
		//	get a minor performance gain at the expense of making the VAX run right!

        if( vax-> quantum.initial ) {

			//	There are some odd cases where quantum is set to the next pending interrupt
			//	quantum boundary.  I haven't found all those yet, so coerce the quantum to
			//	be always be no larger than the initial setting.

			if( vax-> quantum.current > vax-> quantum.initial )
				vax-> quantum.current = vax-> quantum.initial;

			//	It's a quantum event (an instruction executed).  So see if it's now time
			//	to rescan for interrupts, etc.

            vax-> quantum.current--;
            if( vax-> quantum.current <= 0 ) {

				//	We need to learn what the minimum new quantum value should be.  Start
				//	by assuming that it will be the initial value by default.

				minquantum = vax-> quantum.initial;
				
				//	We need to see if an interrupt is ready to go.  Note that we don't
				//	do this if there is already an interrupt pending.
		
				if( !vax-> interrupt_pending ) {

					iplast = 0L;
					ipnext = 0L;
					iquantum = vax-> quantum.initial;
					
					//	Scan the queue of pending interrupts to see if one is ready.

					for( ip = vax-> iqueue; ip; ip = ipnext  ) {
						ipnext = ip-> next;

						if( vax-> debug & DBG_INTERRUPTS ) {
							printf( "DEBUG(INTERRUPTS): quantum; evaluating interrupt %ld, IPL %ld, age=%ld, quantum %ld\n",
								ip-> code, ip-> ipl, ip-> age, ip-> quantum );
						}

						//	If the quantum on this interrupt is not yet zero, it hasn't
						//	aged enough to be taken.  Skip to the next one.

						if( ip-> age > 0 ) {
							iplast = ip;
							continue;
						}

						//	If this interrupt is ready (quantum <= 0) and it's IPL is greater
						//	than the current IPL, then we take this interrupt.  In taking the
						//	interrupt, we also remove it from the queue.

						if( ( unsigned long ) ip-> ipl > ( unsigned long ) vax-> pslw.ipl ) {
							vax-> interrupt_pending = ip-> code;
							vax-> interrupt_ipl = ip-> ipl;
							iquantum = ip-> quantum;
							if( ip == vax-> iqueue )
								vax-> iqueue = ipnext;
							else
								iplast-> next = ipnext;

							if( vax-> debug & DBG_INTERRUPTS ) {
								printf( "DEBUG(INTERRUPTS): set interrupt %ld\n",
									vax-> interrupt_pending );
							}

								
							freemem( ip );
							break;
						}
						iplast = ip;
					}
					
					//	Okay, age all the other interrupts accordingly.  We scan the list of
					//	remaining interrupts.  We reduce their quantum by the amount of time
					//	that the previous interrupt was waiting.  We'll also find out what the
					//	smallest amount of time still remaining is, and use that as the new
					//	minimum quantum value to apply.
					
					if( iquantum ) {
						minquantum = 0x7FFFFFFF;
						if(( vax-> iqueue ) != 0L && ( vax-> debug & DBG_INTERRUPTS )) {
							printf( "DEBUG(INTERRUPTS): resetting pending interrupt quantums by %ld\n",
								iquantum );
						}
						
						for( ip = vax-> iqueue; ip; ip = ip-> next ) {


							if( ip-> age > 0 )
								ip-> age = ip-> age - 1;
							else
								ip-> age = 0;

							if( ip-> quantum < minquantum )
								minquantum = ip-> quantum;

							if( vax-> debug & DBG_INTERRUPTS ) {
								printf( "DEBUG(INTERRUPTS): requeue interrupt %ld, IPL %ld, age=%ld, quantum %ld\n",
									ip-> code, ip->ipl, ip-> age, ip->quantum );
							}

							}
					}
					else
						minquantum = vax-> quantum.initial;
				}
         
         //	Now that we're done with this, reset the quantum.  This is either the
         //	initial quantum value, or the time of the next-most-likely interrupt 
         //	to occur.
         
                vax-> quantum.current = minquantum;

            }
        }


		//	Whew.  After all that, if we decided to take an interrupt, let's
		//	handle it now.  This will take the interrupt, set it up on the
		//	appropriate stack, and requeue the PC appropriately.

		if( vax-> interrupt_pending ) {

			set_fault( vax, vax-> interrupt_pending, 0 );

			// Use the IPL register as temp storage of NEW IPL.  This is the
			// only time IPL and PSL<IPL> are allowed to differ.  The handle
			// fault routine will use the IPL register value to store the
			// NEW IPL on return from this routine.

			vax-> IPL = vax-> interrupt_ipl;
			
			vax-> interrupt_pending = 0;
			vax-> interrupt_ipl = 0;
			
			for( brkp = vax-> console.breakpoint_list; brkp; brkp = brkp -> next ) {
				if(( brkp-> kind == BREAK_FAULT ) && 
				   ( brkp-> pc == vax-> fault.code )) {
					vax-> fault_pending = vax-> fault.code;
					vax-> halted = 0;
					return VAX_BREAK;
				}
			}
			
			handle_fault( vax );
		}
		
		
        //  Move the trace pending bit to the trace bit.  This can have been set by
        //  a previous instruction.
        
        vax-> pslw.tp = vax-> pslw.t;
        
        //  See if the PC value is a stored break point location.  We don't
        //  do this if it's the PC we started at.  If it's the initial PC
        //  then reset our stored memory of the initial PC to an impossible
        //  value, so subsequent hits on this PC will cause a brkp.

        if( vax-> PC == initial_PC )
            initial_PC = 0xFFFFFFFF;
        else
            for( brkp = vax-> console.breakpoint_list; brkp; brkp = brkp-> next ) {
                if(( brkp-> kind == BREAK_INSTRUCTION ) && 
                   ( brkp-> pc == vax-> PC )) {

					if( brkp-> after ) {
						brkp-> after--;
						continue;
					}

                    vax-> halted = 0;
                    format_instruction( vax, vax-> PC );
                    return VAX_BREAK;
                }
            }
        
        
        //  Decode the instruction (and optionally disassemble it)
        
        saved_pc = vax-> PC;
        rc = decode_instruction( vax, &opcode, disasm );
        if( rc ) {
            vax-> PC = saved_pc;
			for( brkp = vax-> console.breakpoint_list; brkp; brkp = brkp -> next ) {
				if(( brkp-> kind == BREAK_FAULT ) && 
				   ( brkp-> pc == vax-> fault.code )) {
					vax-> fault_pending = vax-> fault.code;
					return VAX_BREAK;
				}
			}
            rc = handle_fault( vax );
            if( rc )
                return rc;
            continue;
        }
        
        //  If we're disassembling, print it out now.
                    
        if( disasm ) {
            if( vax-> pslw.is )
                modep = "ISP";
            else
                modep = mode_name[ vax-> pslw.cur_mod ];

            printf( "[%s %08lX] %s\n", 
                 modep, vax-> SP, opcode.name );

        }

        //  Fetch the emulator routine for this instruction.  If it's
        //  not filled in, then error!  Signal reserved/privileged
        //  fault and try to handle from there.
        
        routine = instruction[ opcode.index ].routine;
        if( routine == 0L ) {
            vax-> PC = saved_pc;
            set_fault( vax, EXC_PRIV, 0 );
            rc = handle_fault( vax );
            if( rc )
                return rc;
            continue;
        }
        
        //  Initialize the temp register index freshly
        
        vax-> treg = MAXREG;
        
        //  Update the count of instructions executed.  This is for the TIME
        //  console function.
        
        vax-> console.instruction_count++;
        
        //  Call the emulator routine to handle the opcode.
        
        rc = (*routine)( vax, &opcode );
    
        
        //  The vax already has the fault data stored away, so just
        //  handle it.
        
        if( rc == VAX_FAULT ) {
            vax-> PC = vax-> instruction_PC;
			for( brkp = vax-> console.breakpoint_list; brkp; brkp = brkp -> next ) {
				if(( brkp-> kind == BREAK_FAULT ) && 
				   ( brkp-> pc == vax-> fault.code )) {
					vax-> fault_pending = vax-> fault.code;
					return VAX_BREAK;
				}
			}
            rc = handle_fault( vax );
            if( rc )
                return rc;
                continue;
        }

        //  If after all that, there's some outstanding error, return it.
                
        if( rc )
            return rc;
    
        //  If the trace  bit was set, then after executing the instruction, we
        //  clear the bit and stop execution, without affecting the HALTED state
        //  of the cpu

        if( vax-> pslw.tp ) {
            vax-> pslw.tp = 0;
            rc = set_fault( vax, EXC_TP, 0 );
            rc = handle_fault( vax );
            if( rc )
                return rc;
            continue;
        }
        
        //  If the console is single stepping, then let's stop now
        
        if( vax-> console.singlestep ) {
            vax-> halted = 1;
            vax-> console.singlestep = 0;
            return VAX_OK;
        }
            
    }
    
    return vax-> halted;	/* Reason code stored here by whomever does the halt! */
}

